/*
 * Copyright (c) University of Zagreb, Faculty of Electrical Engineering and Computing
 * Authors: Nikola Banic <nikola.banic@fer.hr> and Sven Loncaric <sven.loncaric@fer.hr>
 * 
 * This is only a research code and is therefore only of prototype quality.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * LITERATURE:
 * 
 * N. Banic and S. Loncaric
 * "Firefly: A Hardware-Friendly Real-Time Local Brightness Adjustment Method"
 * ICIP 2015
 */

#include <cstdio>
#include <cmath>

#if defined(_WIN32) || defined(_WIN64)
#include <cv.h>
#include <highgui.h>
#else
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#endif

#define CV_LOAD_IMAGE_ANYDEPTH 2
#define CV_LOAD_IMAGE_ANYCOLOR 4
#define COLOR_BGR2GRAY 6

class Firefly{
public:
	Firefly();
	~Firefly();
	int* GetLUT();
	void GetLUT(int *lut);
	void SetLUT(int *lut);
	int GetSprayPoolSize();
	void SetSprayPoolSize(int sprayPoolSize);
	int GetSpread();
	void SetSpread(int spread);
	int GetRefine();
	void SetRefine(int refine);
	void Adjust(cv::Mat source, cv::Mat &destination);
	static void CalculateMean(cv::Mat source, cv::Mat &destination, int sprayPoolSize, int spread, int refine);
private:
	int lut[256*256];
	int sprayPoolSize;
	int spread;
	int refine;
};

Firefly::Firefly(){
	SetSprayPoolSize(30);
	SetSpread(5);
	SetRefine(1);
}

Firefly::~Firefly(){}

int* Firefly::GetLUT(){
	return lut;
}

void Firefly::GetLUT(int *lut){
	for (int i=0;i<256*256;++i){
		lut[i]=this->lut[i];
	}
}

void Firefly::SetLUT(int *lut){
	for (int i=0;i<256*256;++i){
		this->lut[i]=lut[i];
	}
}

void Firefly::Adjust(cv::Mat source, cv::Mat &destination){

	cv::Mat result;
	cv::Mat gray;
	source.convertTo(gray, CV_8U);
	source.convertTo(result, CV_64F);
	int channelsCount=source.channels();
	if (gray.channels()>1){
		cvtColor(gray, gray, COLOR_BGR2GRAY);
	}
	gray.convertTo(gray, CV_64F);

	int rows=gray.rows;
	int cols=gray.cols;

	cv::Mat smoothed;
	CalculateMean(gray, smoothed, GetSprayPoolSize(), GetSpread(), GetRefine());
	double *parameterData=(double *)(smoothed.data);

	double upperBound=255;
	
	double *data=(double *)(gray.data);
	for (int i=0;i<rows;++i){
		for (int j=0;j<cols;++j){
			double grayOriginal=*(data+i*cols+j);
			double parameter=*(parameterData+i*cols+j);
			double grayNew=lut[(int)(256*(int)grayOriginal+(int)parameter)];
			double factor=1.0;

			if (grayOriginal==0){
				factor=1.0;
			} else{
				factor=grayNew/grayOriginal;
			}
			
			if (channelsCount==1){
				if (factor*grayOriginal>upperBound){
					*(((double *)(result.data))+i*cols+j)=upperBound;
				}
			} else if (channelsCount==3){
				cv::Vec3d &value=*(((cv::Vec3d *)(result.data))+i*cols+j);
				for (int ch=0;ch<3;++ch){
					if (factor*value[ch]>upperBound){
						factor=upperBound/value[ch];
					}
				}
				for (int ch=0;ch<3;++ch){
					value[ch]*=factor;
				}
			}
		}
	}

	result.convertTo(destination, source.type());
}

void Firefly::CalculateMean(cv::Mat source, cv::Mat &destination, int sprayPoolSize, int spread, int refine){
	cv::Mat img;

	source.convertTo(img, CV_64F);
	int channelsCount=img.channels();
	if (channelsCount>1){
		img.convertTo(img, CV_32F);
		cvtColor(img, img, COLOR_BGR2GRAY);
		img.convertTo(img, CV_64F);
	}
	int rows=img.rows;
	int cols=img.cols;
	
	cv::Mat result(rows, cols, CV_64FC1);
	
	double *integral=new double[sprayPoolSize*sprayPoolSize];
	double *average=new double[sprayPoolSize*sprayPoolSize];

	double *data=(double *)(img.data);
	
	for (int i=0;i<sprayPoolSize;++i){
		for (int j=0;j<sprayPoolSize;++j){
			int cr=i*(rows-1)/(sprayPoolSize-1);
			int cc=j*(cols-1)/(sprayPoolSize-1);
					
			integral[i*sprayPoolSize+j]=*(data+cr*cols+cc);

			if (i>0 && j>0){
				integral[i*sprayPoolSize+j]+=integral[(i-1)*sprayPoolSize+j]+integral[i*sprayPoolSize+(j-1)]-integral[(i-1)*sprayPoolSize+(j-1)];
			} else if(i>0){
				integral[i*sprayPoolSize+j]+=integral[(i-1)*sprayPoolSize+j];
			} else if(j>0){
				integral[i*sprayPoolSize+j]+=integral[i*sprayPoolSize+(j-1)];
			}
		}
	}

	for (int i=0;i<sprayPoolSize;++i){
		int i1=i-spread-1;
		int i2=i+spread;
		if (i1<0){
			i1=0;
		}
		if (i2>=sprayPoolSize){
			i2=sprayPoolSize-1;
		}
		for (int j=0;j<sprayPoolSize;++j){
			int j1=j-spread-1;
			int j2=j+spread;
			if (j1<0){
				j1=0;
			}
			if (j2>=sprayPoolSize){
				j2=sprayPoolSize-1;
			}
			average[i*sprayPoolSize+j]=integral[i2*sprayPoolSize+j2]-integral[i2*sprayPoolSize+j1]-integral[i1*sprayPoolSize+j2]+integral[i1*sprayPoolSize+j1];
			average[i*sprayPoolSize+j]/=((i2-i1)*(j2-j1));
		}
	}

	for (int r=0;r<refine;++r){
		for (int i=0;i<sprayPoolSize;++i){
			for (int j=0;j<sprayPoolSize;++j){
				integral[i*sprayPoolSize+j]=average[i*sprayPoolSize+j];

				if (i>0 && j>0){
					integral[i*sprayPoolSize+j]+=integral[(i-1)*sprayPoolSize+j]+integral[i*sprayPoolSize+(j-1)]-integral[(i-1)*sprayPoolSize+(j-1)];
				} else if(i>0){
					integral[i*sprayPoolSize+j]+=integral[(i-1)*sprayPoolSize+j];
				} else if(j>0){
					integral[i*sprayPoolSize+j]+=integral[i*sprayPoolSize+(j-1)];
				}
			}
		}

		for (int i=0;i<sprayPoolSize;++i){
			int i1=i-spread-1;
			int i2=i+spread;
			if (i1<0){
				i1=0;
			}
			if (i2>=sprayPoolSize){
				i2=sprayPoolSize-1;
			}
			for (int j=0;j<sprayPoolSize;++j){
				int j1=j-spread-1;
				int j2=j+spread;
				if (j1<0){
					j1=0;
				}
				if (j2>=sprayPoolSize){
					j2=sprayPoolSize-1;
				}
				average[i*sprayPoolSize+j]=integral[i2*sprayPoolSize+j2]-integral[i2*sprayPoolSize+j1]-integral[i1*sprayPoolSize+j2]+integral[i1*sprayPoolSize+j1];
				average[i*sprayPoolSize+j]/=((i2-i1)*(j2-j1));
			}
		}
	}

	for (int row=0;row<rows;++row){
		for (int col=0;col<cols;++col){
			
			int resolutionRow=(sprayPoolSize-1)*row/(rows-1);
			int resolutionCol=(sprayPoolSize-1)*col/(cols-1);

			double current=0.0;
			
			int r1=resolutionRow;
			int r2=resolutionRow+1;
			int c1=resolutionCol;
			int c2=resolutionCol+1;

			if (r2>=sprayPoolSize){
				r2=r1;
			}
			if (c2>=sprayPoolSize){
				c2=c1;
			}

			int cr=r1*(rows-1)/(sprayPoolSize-1);
			int cc=c1*(cols-1)/(sprayPoolSize-1);
			int cr2=r2*(rows-1)/(sprayPoolSize-1);
			int cc2=c2*(cols-1)/(sprayPoolSize-1);

			double v00=average[r1*sprayPoolSize+c1];
			double v01=average[r1*sprayPoolSize+c2];
			double v10=average[r2*sprayPoolSize+c1];
			double v11=average[r2*sprayPoolSize+c2];
					
			double w00=(cr2-row)*(cc2-col);
			double w01=(cr2-row)*(col-cc);
			double w10=(row-cr)*(cc2-col);
			double w11=(row-cr)*(col-cc);

			if (w00+w01+w10+w11==0){
				current=average[r1*sprayPoolSize+c1];
			} else{
				current=w00*v00+w01*v01+w10*v10+w11*v11;
				current/=(w00+w01+w10+w11);
			}

			*(((double *)(result.data))+row*cols+col)=current;

		}
	}
	
	delete[] integral;
	delete[] average;

	int type=source.type();
	result.copyTo(destination);
	if (destination.type()!=type){
		destination.convertTo(destination, type);
	}
	
}

int Firefly::GetSprayPoolSize(){
	return sprayPoolSize;
}

void Firefly::SetSprayPoolSize(int sprayPoolSize){
	this->sprayPoolSize=sprayPoolSize;
}

int Firefly::GetSpread(){
	return spread;
}

void Firefly::SetSpread(int spread){
	this->spread=spread;
}

int Firefly::GetRefine(){
	return refine;
}

void Firefly::SetRefine(int refine){
	this->refine=refine;
}

void LoadLUT(const char *fileName, int *lut){
	
	FILE *input=fopen(fileName, "rt");
	for (int i=0;i<256;++i){
		for (int j=0;j<256;++j){
			fscanf(input, "%d", &lut[i*256+j]);
		}
	}
	fclose(input);
	
}

#define LINE_LENGTH 1024
char inputLine[LINE_LENGTH+1];

int main(int argc, char **argv){

	if (argc<4){
		printf("Usage: %s input_file output_file configuration_file\n", argv[0]);
		return 0;
	}

	cv::Mat img=cv::imread(argv[1], CV_LOAD_IMAGE_ANYDEPTH|CV_LOAD_IMAGE_ANYCOLOR);
	
	FILE *input=fopen(argv[3], "rt");
	int sprayPoolSize;
	fgets(inputLine, LINE_LENGTH, input);
	sscanf(inputLine, "%d", &sprayPoolSize);
	int spread;
	fgets(inputLine, LINE_LENGTH, input);
	sscanf(inputLine, "%d", &spread);
	int refine;
	fgets(inputLine, LINE_LENGTH, input);
	sscanf(inputLine, "%d", &refine);
	
	fgets(inputLine, LINE_LENGTH, input);
	while(inputLine[strlen(inputLine)-1]=='\n' || inputLine[strlen(inputLine)-1]=='\r'){
		inputLine[strlen(inputLine)-1]='\0';
	}
	int lut[256*256];
	LoadLUT(inputLine, lut);
	
	Firefly firefly;
	firefly.SetSprayPoolSize(sprayPoolSize);
	firefly.SetSpread(spread);
	firefly.SetRefine(refine);
	firefly.SetLUT(lut);
	
	cv::Mat result;
	firefly.Adjust(img, result);
	
	cv::imwrite(argv[2], result);

	return 0;
}
